//  Copyright (c) 1997-2004 by Mark F. Adams 
//  $Id: agg_grid.C,v 1.39 2005/05/19 19:55:17 adams Exp $
//  Author:             Mark F. Adams
//  Filename:           agg_grid.C
// 
#include "prometheus.hh"

// lapack routines (FORTRAN)
#if defined(__cplusplus)
extern "C"{
#endif
  void *dgeqrf_(int*, int*, double*, int*, double*, double*, int*, int*);
  void *dorgqr_(int*,int*,int*,double*, int*, double*, double*, int*,int*);
#if defined(__cplusplus)
} 
#endif

//! fine PromAggGrid constructor
PromAggGrid::PromAggGrid(const PromOptions &opts,PromPerfMonitor &perf,
			 PromParentBank * const bank, const int nloc,
			 const PromTable<int> &gid_ghostLid,
			 const PromList<int*> *elemList,
			 const PromComm &wcomm, const int ndf ) :
  PromGrid( opts, perf, bank, nloc, wcomm, ndf ), amg_factor_(1), amg_smooths_(0)
{
  for(int jj=0;jj<PROM_MAXNDF;jj++) A_RBMs_[jj] = NULL;
  int ierr = InitNodes( nloc + gid_ghostLid.getCount() );
  if( ierr == 0 ) ierr = FineInit( NULL, ndf, gid_ghostLid, elemList );
  if( ierr == 0 ) ierr = SetAggMethod();

  assert(ierr==0);
}
 
//! normal coarse grid (const dof per node)
PromAggGrid::PromAggGrid(const PromOptions &opts, PromPerfMonitor &perf, 
			 PromAggGrid *lastgrid, const int nloc, 
			 const int nnodes, const PromComm &wcomm, 
			 const int ndf ):
  PromGrid(opts, perf, lastgrid->parBank_, nloc, wcomm, ndf ) , 
  amg_factor_( lastgrid->amg_factor_), amg_smooths_(lastgrid->amg_smooths_)
{
  for(int jj=0;jj<PROM_MAXNDF;jj++) A_RBMs_[jj] = NULL;
  int ierr = InitNodes( nnodes );
  if( ierr == 0 ) ierr = InitNormal( lastgrid );
  assert(ierr==0);
}

//! dummy coarse grid
PromAggGrid::PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
			  PromAggGrid *lastgrid, const int fact, 
			  const PromComm &wcomm, const int bsize ) :
  PromGrid( opts, perf, lastgrid->parBank_, 0, fact, wcomm, bsize ),
  amg_factor_( lastgrid->amg_factor_ ), amg_smooths_( lastgrid->amg_smooths_ )
{
  for(int jj=0;jj<PROM_MAXNDF;jj++) A_RBMs_[jj] = NULL;
  int ierr = InitNodes( 0 );
  if( ierr == 0 ) ierr = InitDummy( lastgrid );
  assert(ierr==0);
}

//! reduced coarse grid, splitter, (const dof per node)
PromAggGrid::PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
			  PromAggGrid *lastgrid, const int factor, 
			  const int buffsz, int *buffer, const int rtag, 
			  PromTable<int> &gid1_lid, 
			  PromList<PromMPINode*> &ghostlist, 
			  const int partloc, const int nnodes,
			  const PromComm &wcomm, const int bsize ) :
  PromGrid(opts, perf, lastgrid->parBank_, partloc, factor, wcomm,bsize),
  amg_factor_( lastgrid->amg_factor_ ), amg_smooths_( lastgrid->amg_smooths_ )
{
  for(int jj=0;jj<PROM_MAXNDF;jj++) A_RBMs_[jj] = NULL;
  int ierr = InitNodes( nnodes );
  assert( lastgrid && lastgrid->isActive() );
  if(ierr == 0) ierr = InitReduced( lastgrid, factor, buffsz, buffer, rtag,
				    gid1_lid, ghostlist );
  assert(ierr==0);
}

//! reduced coarse grid
PromAggGrid::PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
			  PromAggGrid *lastgrid, const int factor, 
			  const int buffsz, int *buffer, const int rtag, 
			  PromTable<int> &gid1_lid,
			  PromList<PromMPINode*> &ghostlist, 
			  const int partloc, const int nnodes,
			  const int nLocPartEq, const PromComm &wcomm ) :
  PromGrid(opts,perf,lastgrid->parBank_,partloc,factor,nLocPartEq,wcomm),
  amg_factor_( lastgrid->amg_factor_ ), amg_smooths_(lastgrid->amg_smooths_) 
{
  for(int jj=0;jj<PROM_MAXNDF;jj++) A_RBMs_[jj] = NULL;
  int ierr = InitNodes( nnodes );
  assert( lastgrid && lastgrid->isActive() );
  if(ierr == 0) ierr = InitReduced( lastgrid, factor, buffsz, buffer, rtag,
				    gid1_lid, ghostlist );
  if(ierr == 0 ) { // fix VBR hack
    int xx,*lid_ndf;ierr = PetscMalloc((nNodes_+1)*sizeof(int),&lid_ndf);
    for(xx=0;xx<nNodes_;xx++)lid_ndf[xx] = nodes_[xx]->getNDF();
    SetLidLeq( lid_ndf );
    PetscFree( lid_ndf );
  }
  assert(ierr==0);
}

//! fine PromAggGrid constructor
PromAggGrid::PromAggGrid(const PromOptions &opts,PromPerfMonitor &perf,
			 PromParentBank * const bank, const int nloc,
			 const PromTable<int> &gid_ghostLid,
			 const PromList<int*> *elemList,
			 const PromComm &wcomm, const int lid_ndof[] ) :
  PromGrid( opts, perf, bank, nloc, wcomm, lid_ndof ),
  amg_factor_(1), amg_smooths_(0)  
{
  for(int jj=0;jj<PROM_MAXNDF;jj++) A_RBMs_[jj] = NULL;
  int ierr = InitNodes( nloc + gid_ghostLid.getCount() );
  if( ierr == 0 ) ierr = FineInit( lid_ndof, -1, gid_ghostLid, elemList );
  if( ierr == 0 ) ierr = SetAggMethod();
  assert(ierr==0);
}

//! normal coarse grid
PromAggGrid::PromAggGrid(const PromOptions &opts, PromPerfMonitor &perf,
			 PromAggGrid *lastgrid, const int nloc, 
			 const int nnodes, const PromComm &wcomm, 
			 const int lid_ndof[] ) :
  PromGrid( opts, perf, lastgrid->parBank_, nloc, wcomm, lid_ndof ),
  amg_factor_( lastgrid->amg_factor_), amg_smooths_(lastgrid->amg_smooths_)
{
  for(int jj=0;jj<PROM_MAXNDF;jj++) A_RBMs_[jj] = NULL;
  int ierr = InitNodes( nnodes );
  if( ierr == 0 ) ierr = InitNormal( lastgrid );
  assert(ierr==0);
}

//! dummy coarse grid
PromAggGrid::PromAggGrid( const PromOptions &opts, PromPerfMonitor &perf, 
			  PromAggGrid *lastgrid, const int fact, 
			  const PromComm &wcomm ) :
  PromGrid(opts, perf, lastgrid->parBank_, 0, fact, wcomm, 0 ),
  amg_factor_( lastgrid->amg_factor_ ), amg_smooths_( lastgrid->amg_smooths_ )
{
  for(int jj=0;jj<PROM_MAXNDF;jj++) A_RBMs_[jj] = NULL;
  int ierr = InitNodes( 0 );
  if( ierr == 0 ) ierr = InitDummy( lastgrid );
  assert(ierr==0);
}

//! destructor
PromAggGrid::~PromAggGrid(){
  for(int i=0;i<PROM_MAXNDF;i++) if(A_RBMs_[i] != NULL) delete A_RBMs_[i];
}
#undef __FUNCT__
#define __FUNCT__ "PromAggGrid::FineGridInitWithA" 
/* PromAggGrid::FineGridInitWithA ******************************************
 *
 */  
int PromAggGrid::FineGridInitWithA() 
{
  assert(this != NULL);

  PetscFunctionBegin;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PromAggGrid::newNode" 
/* PromAggGrid::newNode ******************************************
 *
 */  
PromNode *PromAggGrid::newNode( const int lid )
{
  PromAggNode *curr = newAggNode();            
  curr->index_ = lid;
  return curr;
}

#undef __FUNCT__
#define __FUNCT__ "PromAggGrid::getMGType" 
/* PromAggGrid::getMGType ******************************************
 *
 */  
PromMGType PromAggGrid::getMGType()
{ 
  return PROM_AGGMG; 
}

#undef __FUNCT__
#define __FUNCT__ "PromAggGrid::SetAggMethod" 
/* PromAggGrid::SetAggMethod ****************************************
 *
 *  private fine grid constructor
 */  
int PromAggGrid::SetAggMethod() 
{   
  int ierr,ii,xx,kk;

  // amg stuff -- default solver is smoothed aggregation
  if( isBase() ) {
    bool flag; 
    ierr = options_.HasOptionName("-aggmg_factor", &flag ); CHKERRQ(ierr);
    if( flag ){
      amg_factor_ = 1; // global is default
      ierr = options_.GetOptionInt("-aggmg_factor", &ii, &flag);CHKERRQ(ierr);
      if( flag ) amg_factor_ = ii; // factor to reduce at each level
    } 
    else amg_factor_ = 1; // global is default  
    ierr = options_.GetOptionInt("-aggmg_smooths", &ii, &flag); CHKERRQ(ierr);
    if( flag ) amg_smooths_ = ii;
    else amg_smooths_ = 0; // no smoothing is default
  }
  assert( amg_factor_ != 0 );
  // set heuristic for amg here - if negative then dynamic (positive) value
  if( amg_factor_ < 0 ){
    for(xx=kk=0;xx<nLocalNd_;xx++)kk+=nodes_[xx]->symbData_->adjac.getCount();
    MPI_Allreduce( &kk, &ii, 1, MPI_INT, MPI_SUM, gComm_ );
    amg_factor_ = ii/nglobal() + 1; 
    if( options_.verbose_>1 ) 
      PetscPrintf(gComm_, "\t[%d]%s(%d) new amg factor: %d\n",
		 comm_.mype(), __FUNCT__, getLevel(), amg_factor_);
  }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::CreateCoarseGridSpace" 
/* PromAggGrid::CreateCoarseGridSpace ***************************
 *
 *  - Cosrse grid method: take lasr grid's aggregated nodes and create 
 *      coarse grid space   (needs to coasen to predict next size)
 *
 *  INPUT:
 *    - nextGrid: coarse grid
 *
 *  SIDE EFFECTS:
 *    - make "parents"
 *    - set initial Rx1
 *
 *  RETURN: 
 *    - PET cerror code  
 */
int PromAggGrid::CreateCoarseGridSpace( PromGrid *lgrid )
{
  PromAggGrid *lastGrid = dynamic_cast<PromAggGrid*>(lgrid);
  assert(lastGrid!=NULL && this!=NULL);
  int ierr, xx, mype=comm_.mype(); 
  PromAggNode *fine; assert( lastGrid->isActive() ); 

  // alloc space, set last parents up to here -- marks covered nodes
  ierr = lastGrid->AllocPar0( this ); CHKERRQ(ierr);

  // QR: (this prolongator)(next RBM) = (this RBM) 
  ierr = QRShapes( lastGrid );            CHKERRQ(ierr);

  // clear marks and Bs
  for( xx = 0 ; xx < lastGrid->nNodes_ ; xx++ ){
    fine = dynamic_cast<PromAggNode*>(lastGrid->nodes_[xx]);
    assert(fine && !fine->isDead());
    fine->mark(FALSE); 	fine->SetBsz(0); // no need any more
  }

  // setup added parents for smoothing
  if( amg_smooths_ > 0 ){
    // set up data structures
    if(options_.verbose_>1){
      PetscLogDouble rss; PetscMemoryGetCurrentUsage( &rss ); int lev=getLevel();
      PetscPrintf(lastGrid->gComm_,"[%d]%s MakeFirstParentLayer(%d) %d Mb\n",
		  mype,__FUNCT__,lev,(int)(rss/1.e6+.5));
    }

    // setup data structures for smoothing
    ierr = lastGrid->MakeFirstParentLayer();            CHKERRQ(ierr);

    // add final adjacs (new ghosts)
    if(options_.verbose_>1) 
      PetscPrintf( lastGrid->gComm_, "[%d]%s AddNextAdjacs(%d)\n",
		   mype, __FUNCT__,getLevel());
    ierr = lastGrid->AddNextAdjacs( this );      CHKERRQ(ierr);
  }

  // count parents
  bool determinate,factorflag; 
  options_.HasOptionName("-prometheus_use_deterministic",&determinate);
  options_.GetOptionInt("-aggmg_factor", &xx, &factorflag );
  if( options_.verbose_> 1 || determinate || (factorflag && xx < 0) ) {
    // diagnostics
    int npar[2],sum[2],factorsize=xx;
    for( xx = npar[0] = 0 ; xx < nLocalNd_ ; xx++ ){
      npar[0] += nodes_[xx]->symbData_->adjac.getCount() + 1;
    }
    if( !isActive() ) npar[0] = 0;
    int lim = determinate ? lastGrid->nNodes_ : lastGrid->nLocalNd_ ;
    for( xx = npar[1] = 0 ; xx < lim ; xx++ ){
      PromNode *curr = lastGrid->nodes_[xx];
      if( xx < lastGrid->nLocalNd_ ) npar[1] += curr->nParents();
      if( determinate ) {
	int ii, np = curr->nParents(); assert(np<=PROM_MAXPAR_SOFT);
	int ids[2*PROM_MAXPAR_SOFT]; PromParent *parbuff[PROM_MAXPAR_SOFT];
	for( ii = 1 ; ii < np ; ii++ ) { // leave first
	  PromParent *par = curr->parents_->data_[ii];
	  parbuff[ii] = par; ids[2*ii] = par->global(); ids[2*ii+1] = ii;
	}
	qsort(&ids[2],np-1,2*sizeof(int), prom_intcompar); // sort on procs
	for( ii = 1 ; ii < np ; ii++ ) {
	  PromParent * par = parbuff[ids[2*ii+1]];
	  curr->parents_->data_[ii] = par;
	}
      }
    }
    if( !lastGrid->isActive() ) npar[1] = 0;
    MPI_Allreduce( npar, sum, 2, MPI_INT, MPI_SUM, lastGrid->gComm_ ); 
    assert((sum[0]-nglobal())%2 == 0);
    if( options_.verbose_> 1 ) 
      PetscPrintf(lastGrid->gComm_,
		  "[%d]%s %d non-zero cells, %d parents tot, %g parents ave\n",
		  mype,__FUNCT__,sum[0],sum[1],
		  (float)sum[1]/(float)lastGrid->nglobal());
    // dynamic amg factor  
    if(factorflag && factorsize < 0) { // dynameic factor based on adjac count
      amg_factor_ = sum[0] / nglobal() + 1;
      if(options_.verbose_>1)
	PetscPrintf(lastGrid->gComm_,"[%d]%s(%d) new amg factor:%d\n",
		    mype, __FUNCT__, getLevel(), amg_factor_ );
    }
    else amg_factor_ = lastGrid->amg_factor_;
  }
  else amg_factor_ = lastGrid->amg_factor_;

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::AllocPar0" 
/* PromAggGrid::AllocPar0 ********************************
 *
 *  - allocate space for rows of interpolation operator
 *  - set parents
 *
 *  INPUT: 
 *    - nextmaster:
 *
 *  SIDE EFFECTS:
 *    - marks "covered" nodes
 *
 *  RETURN:
 *    - PET cerror code 
 */
int PromAggGrid::AllocPar0( PromAggGrid *nnext )
{
  int       ierr,ii,xx,proc,tt,torecv,ndf; 
  const int mype=comm_.mype(),nextmaster=nnext->masterProc(mype);
  const int tag=getNewTag(),tag1=getNewTag();
  const int gtag=nnext->isActive();
  PromNode *c2, *curr; PromAggNode *c3, *f2, *fine;  
  shpfloat *R,*array,temp;    xx = sizeof(temp);  tt = sizeof(MPI_Request); 
  const int fisz = xx/sizeof(int);
  const int reqsz = (tt/xx + ((tt%xx == 0) ? 0 : 1)) * fisz;
  MPI_Status status;  Position pos; 
  PromGlobalID        *dep; 
  PromList<PromAggNode*> *list; TablePos tpos;
  PromList<MPI_Request*> smessages;
  PromTable<PromList<PromAggNode*>*> proc_list(59);

  // set parents, count recvs, collect send nodes
  for( xx = torecv = 0 ; xx < nNodes_ ; xx++){
    fine = dynamic_cast<PromAggNode*>(nodes_[xx]);  assert(!fine->isDead());
    if( fine->getState() == PROM_SELECTED && !fine->isGhost() ) {
      curr = fine->symbData_->parent; assert(curr!=NULL);
      // local selected - recv adj
      if( fine->isBound() ) torecv += fine->symbData_->dependents->getCount();
      const int cndf = curr->getNDF(); assert(cndf > 0);
      for( f2 = fine, tt = 0 ; f2 != NULL ; 
	   f2 = dynamic_cast<PromAggNode*>(f2->getChild()) ) tt += f2->getNDF();
      ierr = PetscMalloc((tt*cndf+1)*sizeof(shpfloat),&array);
      CHKERRQ(ierr); array[tt*cndf] = -6.; R = array;
      // set parents (locals only)
      ierr = fine->SetnParents(parBank_,1);  CHKERRQ(ierr); 
      PromParent *par = fine->parents_->data_[0];
      par->Copy( curr, mype ); // copy some node data to parent
      ndf = fine->getNDF();
      ierr = par->SetP( ndf*cndf, array );  CHKERRQ(ierr); 
      array += ndf*cndf; // pointer to current block in 'B'
      par->setMalloced();
      par->setGlobal( curr->index_.global() );
      par->setProc( nextmaster );
      fine->mark(); 
      for( f2 = dynamic_cast<PromAggNode*>(fine->getChild()) ; f2 != NULL ; 
	   f2 = dynamic_cast<PromAggNode*>(f2->getChild()) ){
	assert(f2->nParents() == 0); 
	ierr = f2->SetnParents(parBank_,1); CHKERRQ(ierr);
	par = f2->parents_->data_[0];
	par->Copy( curr, mype ); // copy some node data to parent
	ndf = f2->getNDF(); 
	ierr = par->SetP( cndf*ndf, array );  CHKERRQ(ierr); 
	array += cndf*ndf;
	par->setMalloced( FALSE );
	par->setGlobal( curr->index_.global() );
	par->setProc( nextmaster );
	f2->mark();
      }
      assert( array == R + tt*cndf ); assert(R[tt*cndf]==-6.);
    } 
    else if( (fine->getState()==PROM_SELECTED && fine->isGhost()) || 
	      fine->symbData_->parent->isGhost() ) {
      // uncovered: selected ghost, uncovered deleted 
      PromNode *parn = fine->symbData_->parent;
      if( fine->getState() == PROM_DELETED ) {
	c3 = dynamic_cast<PromAggNode*>(parn->symbData_->parent); 
      }
      else c3 = dynamic_cast<PromAggNode*>(parn);
      assert(c3!=NULL);
      const int sz = fine->getNDF() * c3->getNDF();
      proc = c3->isGhost() ? c3->symbData_->ghostID.proc() : nextmaster;
      // ghost selected - send adj
      if( fine->getState() == PROM_SELECTED ){
	//assert(c3->isGhost());
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<PromAggNode*>(83); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->addTail( c3 );
      }
      // set pare[0]
      ierr = fine->SetnParents(parBank_,1);  CHKERRQ(ierr); 
      PromParent *par = fine->parents_->data_[0];
      par->Copy( c3, mype ); // copy some node data to parent
      ierr = par->SetPsz( sz ); CHKERRQ(ierr);
      par->setGlobal( c3->index_.global() );
      par->setProc( proc );
    }       // covered deleted
    else assert(fine->getState()==PROM_DELETED && 
		!fine->symbData_->parent->isGhost());
  }
 
  // collect recv counts
  if( nnext->factor_ != factor_ ) {
    if( nnext->isActive() ) {
      for(xx = mype/factor_+1, ii=1 ; ii < nnext->factor_/factor_; xx++, ii++){
	MPI_Recv( &tt, 1, MPI_INT, xx, tag, gComm_, &status );
	torecv += tt;
      }
    }
    else{ 
      MPI_Send( &torecv, 1, MPI_INT, nextmaster/factor_, tag, gComm_ );
      torecv = 0;
    }
  }

  // isend ghost adjacs
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    tt = 2*list->getCount() + 1; // 2 = gid and sentinal
    pos = list->getHeadPosition(); assert(pos!=NULL);
    while( pos ) {
      f2 = list->getNext( pos ); 
      tt += f2->symbData_->adjac.getCount(); // *gidJ
    }
    MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz+1)*sizeof(int), &sreq ); 
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff; buff[tt-1] = buff[tt] = -99;
    assert((size_t)pb%sizeof(shpfloat) == 0);
    while( !list->isEmpty() ) {
      f2 = list->removeHead(); 
      *pb++ = f2->index_.global();
      tpos = f2->symbData_->adjac.getHeadPosition();  //assert(tpos!=0);
      while( tpos ) { 
	c2 = f2->symbData_->adjac.getNext( tpos );
	*pb++ = c2->index_.global();
      }
      *pb++ = -1;
    }
    *pb++ = -2; assert(tt == pb-buff); assert(buff[tt]==-99); assert(buff[tt-1]!=-99);
    // send(tag) < *< gidI, *gidJ, -1 >, -2 >
    MPI_Isend(buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq );
    delete list; 
  }

  // recieve parents
  int todo = 0, nextmy0 = nnext->proc_gnode_[nextmaster];
  while( torecv ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); proc = status.MPI_SOURCE*factor_;
    int *buff; ierr = PetscMalloc(tt*sizeof(int), &buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag) < *< gidI, *gidJ, -1 >, -2 >
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status );
    while( pb-buff < tt-1 ) {
      torecv--; assert(torecv>=0);
      int gid = *pb++; ii = gid - nextmy0;
      assert( ii>=0 && ii<nnext->nLocalNd_ );
      curr = nnext->nodes_[ii];
      while( (gid = *pb++) != -1 )  {
	ii = gid - nextmy0;
	if( ii>=0 && ii<nnext->nLocalNd_ ) c2 = nnext->nodes_[ii];
	else {
	  ierr = nnext->GetProc( gid, proc ); CHKERRQ(ierr);
	  todo += nnext->addGhost( c2, proc, gid, gid-nextmy0, gtag );
	  assert(c2->getState() == PROM_UNKNOWN); 
	}
	assert(curr!=c2);
	if( curr->symbData_->adjac.add(c2) ) c2->symbData_->adjac.add(curr);
      }
    }
    assert(*pb++ == -2 && pb-buff == tt);   PetscFree(buff); 
  }
  // free
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag, gComm_, &ii, &status );   assert(!ii);
  // resolve ghosts
  if( nnext->isActive() ) {
    PromTable<int> procTable(59);
    tpos = proc_list.getHeadPosition(); 
    while(tpos){ proc_list.getNext(tpos, &ii); procTable.add(ii); }
    if(options_.verbose_>1 && mype == comm_.np()-1){
      PetscPrintf(PETSC_COMM_SELF,"[%d]%s %d added ghost\n",mype,__FUNCT__,todo);
    }
    ierr = nnext->ReceiveAddGhost( todo, gtag, procTable ); CHKERRQ(ierr);
  } 
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &ii, &status); assert(!ii);
  
  // get RBM for ghosts
  torecv = nNodes_ - nLocalNd_;
  // collect sends
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
    f2 = dynamic_cast<PromAggNode*>(nodes_[xx]);
    // send
    if( f2->symbData_->dependents ) {
      Position pos = f2->symbData_->dependents->getHeadPosition(); 
      while( pos ) { // send to old ghosts
	dep = f2->symbData_->dependents->getNext( pos );   
	proc = dep->proc(ierr);                     CHKERRQ(ierr);
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<PromAggNode*>(100); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->addTail( f2 ); 
      }
    }
  }
  // send to ghosts
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    pos = list->getHeadPosition(); assert(pos!=NULL); tt = 0;
    while( pos ) {
      fine = list->getNext( pos ); 
      const int sz = fine->getBsz();
      tt += 2;
      if( tt%fisz != 0 ) tt++;
      tt += sz*fisz;
    }
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    assert((size_t)pb%sizeof(shpfloat) == 0);
    while( !list->isEmpty() ) {
      fine = dynamic_cast<PromAggNode*>(list->removeHead()); 
      *pb++ = fine->index_.global();
      int sz = *pb++ = fine->getBsz();      
      R = fine->B_; assert(R!=NULL); // 6x6 or 1x1
      if( (pb-buff)%fisz != 0 ) *pb++ = -3;
      for(ii=0 ; ii < sz ; ii++){ temp = R[ii]; PromContext::Write(temp,pb); }
    }
    assert(pb-buff == tt);
    // send(tag1) *< gidD(loc), sz, [-3,] [R] > 
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag1, gComm_, sreq );
    delete list; 
  }

  // recieve ghosts RBM
  while( torecv ) {
    MPI_Probe( MPI_ANY_SOURCE, tag1, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); proc = status.MPI_SOURCE*factor_;
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag1)  *< gidD(ghost), sz, [-3,] [R] > 
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag1, gComm_, &status );
    while( pb-buff < tt ) {
      torecv--; assert(torecv>=0);
      int gidD = *pb++, sz = *pb++; assert(sz>0);
      f2 = dynamic_cast<PromAggNode*>(ghosts_.find(gidD+1)); assert(f2!=NULL);
      if( f2->B_ == NULL ){ ierr = f2->SetBsz(sz); CHKERRQ(ierr); }
      if( (pb-buff)%fisz != 0 ){ int tv = *pb++; assert(tv == -3); }
      for(ii=0 ; ii<sz ; ii++){PromContext::Read(temp,pb); f2->B_[ii]=temp;}
    }
    assert(pb-buff == tt);    PetscFree(buff);
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &ii, &status);assert(!ii); 
  MPI_Iprobe(MPI_ANY_SOURCE, tag1, gComm_, &ii, &status);assert(!ii); 

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::MakeFirstParentLayer" 
/* PromAggGrid::MakeFirstParentLayer ******************************************
 *
 *  - add parent zero of neighbores
 *
 *  INPUT:
 *    - tag: uses tag
 *
 *  SIDE EFFECTS:
 *  RETURN: 
 *     - PETSc error code 
 */
int PromAggGrid::MakeFirstParentLayer()
{
  int ierr,ii,tt,jj,xx,gid,*pb,proc,pp,doit,receive,todo,maxpar=1,nsent=0;
  MPI_Status    status; const int tag = getNewTag();
  PromParent   *parj0;  
  TablePos      tpos; 
  PromNode     *curr, *n1;                              assert(isActive());
  PromTable<PromList<PromNode*>*>  proc_list(59);
  PromList<PromNode*> *list; 
  PromList<PromParent*> parents(10);
  PromList<MPI_Request*> smessages(10); 
  Position      pos;  PromGlobalID *dep;
#if defined(PARCH_ascired)
  const int bsz = 8*4096, bfisz = (bsz-sizeof(MPI_Request))/sizeof(int);
#else
  const int bsz = 64*4096, bfisz = (bsz-sizeof(MPI_Request))/sizeof(int);
#endif
  // add parents of neighbors    
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
    curr = nodes_[xx]; //const int fndf = curr->ndf(); assert(fndf>0);
    // get new parents
    assert(parents.isEmpty());
    parents.addTail( curr->parents_->data_[0] );
    assert(curr->nParents()==1);
    const PromTable<PromNode*> &adj = curr->symbData_->adjac;
    tpos = adj.getHeadPosition(); 
    while( tpos ) { 
      n1 = adj.getNext( tpos );
      // add if new
      parj0 = n1->parents_->data_[0]; assert(parj0!=NULL); 
      pos = parents.getHeadPosition();       doit = TRUE;
      while( pos && doit ) { // send to old ghosts
	PromParent *p = parents.getNext( pos );
	if( p->global() == parj0->global() ) doit = FALSE;
      }
      if( doit ) parents.addTail( parj0 );
    }
    // add new parents
    parents.removeHead(); // original 1st
    if( !parents.isEmpty() ) {
      tt = parents.getCount() + 1;   assert(curr->nParents() == 1);
      if(tt > maxpar) maxpar = tt;
      ierr = curr->SetnParents(parBank_,tt);                 CHKERRQ(ierr);
      // add parents 
      for( pp = 1 ; pp < tt ; pp++ ) { 
	parj0 = parents.removeHead(); 
	PromParent *parI = curr->parents_->data_[pp];
	parI->PromElement::operator=( *parj0 );
	assert(parI->getPsz() == 0);
	parI->setGlobal( parj0->global() );
	parI->setProc( parj0->proc() );
	parI->setGEQ( parj0->geq() );
      }
      assert(parents.isEmpty());
    }
  }
  assert(parents.isEmpty());
  // count deps
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
    curr = nodes_[xx];
    if( curr->isBound() ) { 
      pos = curr->symbData_->dependents->getHeadPosition(); 
      while( pos ) { // send to old ghosts
	dep = curr->symbData_->dependents->getNext( pos );
	if( (list=proc_list.find(dep->proc()+1))==NULL){
	  list = new PromList<PromNode*>(10); CHKNEW(list,1);
	  proc_list.add( list, dep->proc()+1 );
	}
	list->addTail( curr ); list->addTail( (PromNode*)dep );  // hack
      }
    }
  }
  // count ghosts
  for( xx = nLocalNd_, todo = 0 ; xx < nNodes_ ; xx++ ) {
    curr = nodes_[xx];    todo++;
  }

  // send back new parents counts
  PromElement::GetWrite_size(ii); const int psize = 3+ii; // g,pr,E,geq
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--; 
    MPI_Request *sreq; ierr = PetscMalloc(bsz,&sreq);    CHKERRQ(ierr);
    int *buff = (int*)(sreq + 1), count = 0; smessages.addTail( sreq );
    pb = buff; *pb++ = -1; *pb++ = -1;
    while( !list->isEmpty() ) {
      curr = list->removeHead(); 
      dep = (PromGlobalID*)list->removeHead();  // hack
      assert(dep->proc() == proc);
      // send if necc.
      if( (pb-buff) + 2 + (curr->nParents()-1)*psize > bfisz ) {
	buff[0] = count; buff[1] = proc*bfisz + (pb-buff-1);
	// send(tag)<n,proc*bfsz+(sz-1),n*<gid,np,np*<gidI,procI,ELEMJ,geq>>>
	MPI_Isend( buff, pb-buff, MPI_INT, proc/factor_, tag, gComm_, sreq );
	nsent++;
	// reload
	ierr = PetscMalloc( bsz, &sreq );    CHKERRQ(ierr);
	buff = (int*)(sreq + 1); count = 0; smessages.addTail( sreq );
	pb = buff; *pb++ = -1; *pb++ = -1;
      }
      // 
      count++;
      *pb++ = curr->index_.global(); *pb++ = curr->nParents()-1;
      for( pp = 1 ; pp < curr->nParents() ; pp++ ) {
	PromParent *par = curr->parents_->data_[pp];
	*pb++ = par->global();
	*pb++ = par->proc();
	par->PromElement::Write( &pb );
	*pb++ = par->geq();	assert(par->getNDF()>0);
      }
    }
    assert(buff[0]==-1); buff[0] = count; buff[1] = proc*bfisz + (pb-buff-1); 
    // send(tag) <n,proc*bfsz+(sz-1),n*<gid,npar,npar*<gidI,procI,ELEMJ,geq>>>
    MPI_Isend( buff, pb-buff, MPI_INT, proc/factor_, tag, gComm_, sreq );
    delete list;   assert(pb-buff <= bfisz);
    nsent++;
  }

  // receive counts and post recvs
  PromElement dummy; 
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); assert(tt<bsz);
    int buff[bsz]; proc = status.MPI_SOURCE*factor_; pb = buff; 
    // recv(tag) <n,proc*bfisz+(sz-1),n*<gid,npar,npar*<gidI,procI,ELEMJ,geq>>>
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status );
    ii = *pb++; int sz = *pb++; 
    if(sz/bfisz!=comm_.mype())SETERRQ3(1,"ERROR %d/%d!=%d",sz,bfisz,comm_.mype()); 
    sz = sz%bfisz + 1;
    todo -= ii; assert(todo>=0);
    while( pb-buff < tt ) {
      gid = *pb++; jj = *pb++; 
      curr = ghosts_.find(gid+1);     assert(curr!=NULL);
      while( jj-- ){
	gid = *pb++; proc = *pb++; 
	for( pp = 1, doit = TRUE ; pp < curr->nParents() && doit ; pp++ ) {
	  if( curr->parents_->data_[pp]->global() == gid ) doit = FALSE;
	}
	if( doit ){
	  pp = curr->nParents(); assert(pp>0);
	  ierr = curr->SetnParents(parBank_,pp+1); CHKERRQ(ierr); 
	  PromParent *par = curr->parents_->data_[pp]; 
	  par->PromElement::Read( &pb ); 
	  int geq = *pb++;
	  par->setGlobal( gid );  
	  par->setProc( proc );
	  par->setGEQ( geq );
	  if(pp+1 > maxpar) maxpar = pp+1;
	}
	else{ dummy.PromElement::Read(&pb); *pb++; } 
      }
    }
    assert(pb-buff == tt);
  }
  // clean up
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );     PetscFree( sreq );
  }
  // debug
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &receive, &status); assert(!receive);
  if( options_.verbose_ > 1 ){
    PetscLogDouble rss; PetscMemoryGetCurrentUsage( &rss ); int lev = getLevel();
    PetscPrintf(gComm_,"[%d]%s(%d) maxpar=%d nsent=%d bsz=%d %d Mb\n",
		comm_.mype(),__FUNCT__,lev,maxpar,nsent,bsz,(int)(rss/1.e6+.5));
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::AddNextAdjacs" 
/* PromAggGrid::AddNextAdjacs ***********************
 *
 *  - add parent zero of neighbores, do ghosts, and add adjacs to next grid
 *
 *  INPUT:
 *    - nnext:
 *    - doadjacs:
 *
 *  SIDE EFFECTS:
 *
 *  RETURN: 
 *     - PETSc error code 
 */  
int PromAggGrid::AddNextAdjacs( PromAggGrid *nnext )
{
  //#define PROM_USE_IRECV
  int ierr,ii,xx,proci,procj,pi,todo,tt,vb=options_.verbose_;
  MPI_Status  status;  assert(nnext!=NULL); assert(isActive());
  const int reqsz = sizeof(MPI_Request)/sizeof(int),gtag=nnext->getNewTag();
  const int mype = comm_.mype(),np=comm_.np(),tsz=89,tag=getNewTag();
  const int *np_g=nnext->proc_gnode_,mynx0=np_g[mype],nxNloc=nnext->nLocalNd_;
  Position    pos; PromList<int>  *list; 
  TablePos    tpos; tt = 2*nxNloc+17;
  PromTable<PromTable<int>*> parGID_table(tt);
  PromTable<PromList<int>*> proc_list(59); 
  PromTable<int> *gidadjacI,procTable(59);
  PromList<MPI_Request*>  smessages;
#if defined(PROM_USE_IRECV)
  PromList<MPI_Request*>  rmessages; 
#else
  int todo2;
#endif

  // make new adjacs on nnext
  PromTable<int> tGidJs(91);
  for( xx = 0, todo = 0 ; xx < nLocalNd_ ; xx++ ){
    const PromNode*const curri = nodes_[xx];
    // collect J in 'gidadjJArr'
    const PromTable<PromNode*> &adj = curri->symbData_->adjac;
    const PromNode *cj = curri;
    tpos = adj.getHeadPosition();
    while( 1 ){
      const int nparJ = cj->nParents();
      const PromParent *const*const parJarr = cj->parents_->data_;
      for( int pj = 0 ; pj < nparJ ; pj++ ) {
	const PromParent*const parj = parJarr[pj];
	tGidJs.add( parj->global() + 1 );
      }
      // inc loop
      if( tpos ) cj = adj.getNext( tpos );
      else break;
    }
    // add next adjacs - for I
    const int nparI = curri->nParents();
    for( pi = 0 ; pi < nparI ; pi++ ) {
      const PromParent * const parI = curri->parents_->data_[pi];
      const int gidI = parI->global();
      const int ii = gidI - mynx0; 
      ierr = FindProc( np_g, np, gidI, proci );  CHKERRQ(ierr);
      gidadjacI = NULL;
      tpos = tGidJs.getHeadPosition(); 
      while( tpos ) {  // for J
	const int gidJ = tGidJs.getNext( tpos ) - 1; 
	const int jj = gidJ - mynx0;
	ierr = FindProc( np_g, np, gidJ, procj );  CHKERRQ(ierr);
	//
	if( gidJ == gidI ) continue; // no self adds
	// 
	if( proci == mype ) { // local I
	  PromNode *currJ, *currI = nnext->nodes_[ii];
	  if( procj == mype ) currJ = nnext->nodes_[jj];
	  else{ // ghost J
	    // send(gtag) 
	    todo += nnext->addGhost( currJ, procj, gidJ, 
				     gidJ - nnext->proc_gnode_[procj], gtag);
	    procTable.add( procj+1 );
	  }
	  if( currI->symbData_->adjac.add( currJ ) ) {
	    int add = currJ->symbData_->adjac.add( currI ); assert(add==1);
	  }
	}
	else{ // ghost I
	  // get I array 'gidadjacI'
	  if( gidadjacI == NULL ) {
	    if( (gidadjacI=parGID_table.find( gidI+1 )) == NULL ) {
	      gidadjacI = new PromTable<int>(tsz); CHKNEW(gidadjacI,1);
	      int add = parGID_table.add( gidadjacI, gidI+1 ); assert(add);
	    }
	  }
	  // if add I-->J, add J-->I
	  if( gidadjacI->add( gidJ + 1 ) ){
	    if( procj == mype ) { // local J
	      PromNode *currI, *currJ = nnext->nodes_[jj];
	      // send(gtag) for ghost i
	      ierr = FindProc( np_g, np, gidI, proci );  CHKERRQ(ierr);
	      todo += nnext->addGhost(currI, proci, gidI, 
				      gidI - nnext->proc_gnode_[proci],gtag);
	      procTable.add( proci+1 );
	      // add J --> I
	      currJ->symbData_->adjac.add( currI );
	    }
	    else{ // ghost j
	      PromTable<int> *gidadjacJ = parGID_table.find(gidJ+1);
	      if( gidadjacJ == NULL ){
		gidadjacJ = new PromTable<int>(tsz);  CHKNEW(gidadjacJ,1);
		parGID_table.add( gidadjacJ, gidJ+1 ); 
	      }
	      int add = gidadjacJ->add( gidI + 1 ); assert(add==1);
	    }
	  }
	}
      } // parJ
    } // par I
    tGidJs.removeAll();
  }

  // "send" adjacs
  if(vb>1) PetscPrintf(gComm_,"[%d]%s collect sends\n",mype,__FUNCT__);
  tpos = parGID_table.getHeadPosition(); 
  while( tpos ) { 
    gidadjacI = parGID_table.getNext( tpos, &ii ); const int gidI = ii - 1;
    ierr = FindProc( np_g, np, gidI, proci );  CHKERRQ(ierr);
    assert( proci != mype ); 
    if( (list=proc_list.find(proci + 1)) == NULL ) {
      list = new PromList<int>(100); CHKNEW(list,1);
      proc_list.add( list, proci + 1 );
    }
    list->addTail( gidI+1 );
  }
  {// count sends
    int sbuf[PROM_BSZ], rbuf[PROM_BSZ], npp = np/factor_;
    for( ii = 0 ; ii < npp ; ii++ ) sbuf[ii] = 0;
    tpos = proc_list.getHeadPosition(); 
    while( tpos ) {
      list = proc_list.getNext( tpos, &proci ); proci--; 
#if defined(PROM_USE_IRECV)
      pos = list->getHeadPosition(); assert(pos!=NULL); tt = 0;
      while( pos ) {
	const int gidI = list->getNext( pos ) - 1; // gid
	gidadjacI = parGID_table.find( gidI+1 ); assert(gidadjacI != NULL);
	tt += gidadjacI->getCount() + 2; // gid, *gidJ, -1
      }
      assert(sbuf[proci/factor_] == 0); sbuf[proci/factor_] = tt;
#else
      assert(sbuf[proci/factor_] == 0); sbuf[proci/factor_] = 1; 
#endif
    }
#if defined(PROM_USE_IRECV)
    // global!
    MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, gComm_ );
    // post recv(tag)
    for( ii = 0 ; ii < comm_.np()/factor_ ; ii++ ) {
      if( rbuf[ii] != 0 ) {
	// recv(tag)
	int sz = rbuf[ii], part = ii;
	int *pb; ierr = PetscMalloc((sz+reqsz)*sizeof(int),&pb); CHKERRQ(ierr);
	MPI_Request *rreq = (MPI_Request*)pb; pb += reqsz;
	// irecv(tag) *< oldgid, loclid >
	MPI_Irecv( pb, sz, MPI_INT, part, tag, gComm_, rreq );
	rmessages.addTail( rreq );
      }
    }
#else
    // global!
    for( ii = 0 ; ii < npp ; ii++ ) rbuf[ii] = 1;
    MPI_Reduce_scatter( sbuf, &tt, rbuf, MPI_INT, MPI_SUM, gComm_ );
    todo2 = tt;
#endif
  }
  // isend
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proci ); proci--;
    pos = list->getHeadPosition(); assert(pos!=NULL); tt = 0;
    while( pos ) {
      const int gidI = list->getNext( pos ) - 1; // gid
      gidadjacI = parGID_table.find( gidI+1 ); assert(gidadjacI != NULL);
      tt += gidadjacI->getCount() + 2; // gid, *gidJ, -1
    }
    MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz)*sizeof(int), &sreq ); 
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      const int gidI = list->removeHead() - 1;
      gidadjacI = parGID_table.find(gidI+1); assert(gidadjacI != NULL);
      assert(gidadjacI!=NULL && !gidadjacI->isEmpty());
      *pb++ = gidI;
      while( !gidadjacI->isEmpty() ) *pb++ = gidadjacI->removeHead() - 1; 
      *pb++ = -1;
      delete gidadjacI;  // destructive    
    }
    assert(tt==pb-buff);
    // send(tag) *< gid, *<gidadj>, -1 >
    MPI_Isend( buff, pb-buff, MPI_INT, proci/factor_, tag, gComm_, sreq ); 
    delete list; 
  }
#if defined(PROM_USE_IRECV)
  // recv(tag) *< gidI_loc, *<gidadj>, -1 >
  while( !rmessages.isEmpty() ) {
    MPI_Request *req = rmessages.removeHead(); 
    assert(*req != MPI_REQUEST_NULL);
    // recv(tag) * < oldgid, loclid >
    MPI_Test( req, &ii, &status );
    if( !ii ){ rmessages.addTail( req ); continue; }
    MPI_Get_count( &status, MPI_INT, &tt ); assert(tt > 0);
    int *buff = (int*)req + reqsz, *pb = buff; 
#else
    while( todo2 > 0 ) {
      MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
      MPI_Get_count( &status, MPI_INT, &tt ); 
      int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
      // recv(tag) *< gid, *<gidadj>, -1 >
      MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status );
      todo2--; assert(todo2>=0);
#endif
      while( pb-buff < tt ) {
	const int gidI = *pb++;
	ii = gidI - mynx0; assert(ii >= 0 && ii < nxNloc);
	PromNode *currJ, *currI = nnext->nodes_[ii]; int gidJ;
	while( (gidJ=*pb++) != -1 ){
	  ii = gidJ - mynx0; 
	  if(ii < 0 || ii >= nxNloc ){
	    ierr = FindProc( np_g, np, gidJ, procj );  CHKERRQ(ierr);
	    // send(gtag) 
	    todo += nnext->addGhost( currJ, procj, gidJ, 
				     gidJ - nnext->proc_gnode_[procj], gtag );
	    procTable.add( procj+1 );
	  }
	  else currJ = nnext->nodes_[ii];  
	  currI->symbData_->adjac.add(currJ); 
	  currJ->symbData_->adjac.add(currI);
	}
      }
#if defined(PROM_USE_IRECV)
      assert(tt==pb-buff); PetscFree(req); 
    }
#else
    assert(tt==pb-buff); PetscFree(buff); 
  }
#endif
  // free
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 
  if(options_.verbose_>1)PetscPrintf( gComm_,"[%d]%s %d added ghosts\n",
				       comm_.mype(),__FUNCT__,todo);
  if( nnext->isActive() ) {
    ierr = nnext->ReceiveAddGhost( todo, gtag, procTable ); CHKERRQ(ierr);
  }
  // debug
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &ii, &status); assert(!ii);

  return 0;
}

/* int PromAggGrid::SetCoords() ****************************************
 *
 *  - init. shape fuctions with (6) rigid body modes (R == P0)
 *
 *  INPUT:
 *    - coords: 
 *    - right:
 *    - offsets:
 *
 *  SIDE EFFECTS:
 *    - computes 'B'
 *    - calls 'setBsz()' if necc.
 *
 *  RETURN:
 *    - PETSc error code
 */
#undef __FUNCT__   
#define __FUNCT__ "PromAggGrid::SetCoords" 
int PromAggGrid::SetCoords(const double coords[],
			   const double*const offsets/*=NULL*/)const
{
  int       ierr,xx,ii,jj,eq;   assert(isBase());
  double    d0, d1, d2;

  FILE *file = NULL; // output rigid bofy modes
  if( options_.matrix_ ) {
    char path[128]; MPI_Comm_rank(PETSC_COMM_WORLD, &ii); 
    sprintf(path,"RBM_%d", ii );
    file = fopen(path,"w"); assert(file!=NULL);
  }

  for( xx = 0, eq = 0 ; xx < nNodes_ ; xx++ ) {
    PromAggNode *fine = dynamic_cast<PromAggNode*>(nodes_[xx]);
    if(fine==NULL)SETERRQ4(1,"fine==NULL: %d of %d (nloc=%d) %p",
			   xx,nNodes_,nlocal(),nodes_[xx]);
    const int fndf = fine->getNDF();             assert(fndf>0);
    // hack for cdnf !!!
    const int cndf = (fndf>2 && coords != NULL && fndf<6) ? fndf+3 : fndf;

    const int sz = fndf*cndf;   assert(fine->getBsz() == 0); 
    ierr = fine->SetBsz(sz); CHKERRQ(ierr); 
    // elasticity is always first
    switch( fndf ){
    case 1:  fine->B_[0] = 1.; break;
    case 3:  // fallthrough
    case 4:  // fallthrough
    case 6:  // fallthrough
    case 7:  // fallthrough
      if( coords == NULL ) {
	// [ I | 0 ] -- default/base
	for(ii=0;ii<fndf;ii++) {
	  for(jj=0;jj<cndf;jj++) {
	    fine->B_[ii*cndf+jj] = (ii==jj) ? 1. : 0.;
	  }
	}
      }
      else {
	if( offsets != NULL ){ 
	  d0=offsets[xx*3]; d1 = offsets[xx*3+1]; d2=offsets[xx*3+2];
	}
	else d0 = d1 = d2 = 0.;
	// translational (I 3x3)
	for( ii = 0 ; ii < 3 ; ii++ ){
	  for(jj=0 ; jj < 3 ; jj++) fine->B_[ii*cndf+jj] = (ii==jj) ? 1. : 0.;
	}
	// rotational modes - Q in upper right: [ I | Q ] (3x(cndf-3))
	for( ii = 0 ; ii < 3 ; ii++ ){
	  for( jj = 3 ; jj < 6 ; jj++ ){
	    if( ii == (jj-3) ) fine->B_[ii*cndf + jj] = 0.; // sq. symm. diag.
	    else{
	      int index = ii*cndf + jj;
	      if( ii+jj == 4 ) fine->B_[index] = coords[xx*3+2]+d2;
	      else if( ii+jj==5 ) fine->B_[index]= coords[xx*3+1]+d1;
	      else if( ii+jj==6 ) fine->B_[index]= coords[xx*3]+d0;
	      else fine->B_[index] = 0.;
	    }
	  }
	}
	fine->B_[5] *= -1.; fine->B_[cndf+3] *= -1.; fine->B_[2*cndf+4] *= -1.;
	// ii == 3
	if( fndf > 5 ) {
	  for( ii = 3 ; ii < fndf ; ii++ ){
	    for(jj=0;jj<cndf;jj++) fine->B_[ii*cndf+jj] = (ii==jj) ? 1. : 0.;
	  }
	}
	else if( fndf == 4 ) { // add extras (pressures, etc.)
	  assert(ii==3 && cndf==7);
	  for(jj=0;jj<cndf;jj++) fine->B_[3*cndf+jj] = (6==jj) ? 1. : 0.;
	  for(jj=0;jj<3;jj++) fine->B_[jj*cndf+6] = 0.;
	}
      }
      break;
    case 5:  // fallthrough 
    default: 
      // [ I | 0 ] -- default/base
      for(ii=0;ii<fndf;ii++) {
	for(jj=0;jj<cndf;jj++) {
	  fine->B_[ii*cndf+jj] = (ii==jj) ? 1. : 0.;
	}
      }
    }

    if( file != NULL ) {
      double *vv = fine->B_;
      for( ii = 0 ; ii < fndf ; ii++, eq++ ){
	if( cndf == 1 ) { // hack for accoustic
	  fprintf(file,"%d %d %.14e\n", eq+1, 7, *vv++ );
	}
	else {
	  for( jj = 0 ; jj < cndf ; jj++ ) {
	    fprintf(file,"%d %d %.14e\n", eq+1, jj+1, *vv++ );
	  }
	}
      }
    }
  }
  
  if( file != NULL ) fclose(file);

  return 0;
}


#undef __FUNCT__
#define __FUNCT__ "PromAggGrid::ScaleRBM" 
/* PromAggGrid::ScaleRBM ******************************************
 *
 */  
int PromAggGrid::ScaleRBM( PromVector *right ) const
{
  assert(this != NULL); assert(stiffness_ != NULL);
  int xx,ii,jj,eq,ierr;
  double temp,*scale = NULL,*vv;

  PetscFunctionBegin;
  if( right != NULL ) { // not needed - done in 'SetCoords'
    // scale rbm 
    int factor = stiffness_->isComplex() ? 2 : 1;
    ierr = right->GetArray( &scale );    CHKERRQ(ierr);
    for( xx = 0, eq = 0 ; xx < nLocalNd_ ; xx++ ) {
      PromAggNode *fine = dynamic_cast<PromAggNode*>(nodes_[xx]);
      const int fndf = fine->getNDF();                    assert(fndf>0);
      const int cndf = (jj=fine->getBsz())/fndf;          assert(jj%fndf == 0);
      assert(fine->B_ != NULL); vv = fine->B_;
      for( ii = 0 ; ii < fndf ; ii++, eq += factor ){ 
	temp = 1. / scale[eq];
	for( jj = 0 ; jj < cndf ; jj++, vv++ ) *vv *= temp;
      }
    }
    ierr = right->RestoreArray( &scale );    CHKERRQ(ierr);
  }

  // debug -- cache A_RBMs in vectors
  bool flag;
  options_.HasOptionName("-aggmg_check_rbm",&flag);
  if( flag ){
    const PromMatrix *const A = stiffness_; assert(A!=NULL);
    for(ii=0;ii<PROM_MAXNDF;ii++) {
      A_RBMs_[ii] = A->isComplex() ? new PromComplexVector(A->map_) : 
	new PromVector(A->map_);
      A_RBMs_[ii]->Set( 0.0 );
    }
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
      PromAggNode *fine = dynamic_cast<PromAggNode*>(nodes_[xx]);
      const int fndf = fine->getNDF(), sz = fine->getBsz(); assert(fndf>0);
      const int cndf = sz/fndf; assert(cndf<PROM_MAXNDF && sz%fndf==0);
      if( cndf == 1 ) { // hack for accoustics
	PromVector *vec = A_RBMs_[6];
	for(ii=0,eq=fine->geq();ii<fndf;ii++,eq++) {
	  double temp = fine->B_[ii*cndf+6];
	  ierr=vec->SetValues(1, &eq, &temp, INSERT_VALUES);CHKERRQ(ierr);
	}
      }
      else {
	for(jj=0;jj<cndf;jj++) {
	  PromVector *vec = A_RBMs_[jj];
	  for(ii=0,eq=fine->geq();ii<fndf;ii++,eq++) {
	    double temp = fine->B_[ii*cndf+jj];
	    ierr=vec->SetValues(1, &eq, &temp, INSERT_VALUES);CHKERRQ(ierr);
	  }
	}
      } 
    }
    // get fixed nodes and neighbors
    const int nloc = nLocalNd_, mype = comm_.mype(); 
    bool *isfixed = new bool[nloc]; int ii,xx,jj,eq,ids[PROM_MAXNDF];
    for( xx = 0 ; xx < nloc ; xx++ ) isfixed[xx] = false;
    for( xx = 0 ; xx < nloc ; xx++ ) {
      PromAggNode *fine = dynamic_cast<PromAggNode*>(nodes_[xx]);
      // flag fixed and neighbors
      if( fine->isFixed() ) {
	isfixed[xx] = true;
	TablePos tpos = fine->symbData_->adjac.getHeadPosition();
	while( tpos ) {
	  PromNode *f2 = fine->symbData_->adjac.getNext( tpos ); 
	  if( !f2->isGhost() ) isfixed[f2->index_()] = true;
	}
      }
    }
    // check AB == 0
    MPI_Comm wcomm = comm_();
    double tt,errors[PROM_MAXNDF],rerr[PROM_MAXNDF],zeros[2*PROM_MAXNDF];
    ierr = A->NormInf( &tt ); CHKERRQ(ierr); 
    const double norm = tt;   PromVector *work, *vec;
    work = A->isComplex() ? new PromComplexVector(A->map_) : new PromVector(A->map_);
    for(ii=0;ii<2*PROM_MAXNDF;ii++)zeros[ii]=0.;    
    for(jj=0;jj<PROM_MAXNDF;jj++){
      // work <- A * B
      vec = A_RBMs_[jj];
      ierr = vec->Assembly(); CHKERRQ(ierr);
      ierr = A->Mult( vec, work );    CHKERRQ(ierr);
      delete vec; vec = A_RBMs_[jj] = NULL; 
      for(xx=eq=0;xx<nloc;xx++){
	int ndf = nodes_[xx]->getNDF();
	if( isfixed[xx] ){
	  for(ii=0;ii<ndf;ii++) ids[ii] = eq + ii;
	  ierr = work->SetValues(ndf, ids, zeros, INSERT_VALUES); CHKERRQ(ierr);
	}
	eq += ndf;
      }
      ierr = work->Assembly(); CHKERRQ(ierr);
      ierr = work->NormMax( &tt ); CHKERRQ(ierr); 
      errors[jj] = tt/norm;
    }
    for(xx=ii=0 ; xx < nloc ; xx++) if( isfixed[xx] ) ii++;
    MPI_Allreduce( errors, rerr, PROM_MAXNDF, MPI_DOUBLE, MPI_MAX, wcomm );
    PetscPrintf(wcomm,"\t[%d] |AB[:,?]|/|A|=[%5.1e %5.1e %5.1e %5.1e %5.1e %5.1e %5.1e] %3.1g%% BC (|A|=%g)\n",mype,rerr[0],rerr[1],rerr[2],rerr[3],rerr[4],
		rerr[5],rerr[6],100.*(double)ii/(double)nloc,norm);
    delete [] isfixed;    delete work;
  } 
  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "PromAggGrid::QRShapes" 
/* PromAggGrid::QRShapes ***********************************************
 *
 *  - Orthigonalize (Q) on 'lastgrids' (rigid body modes)
 *
 *  INPUT:
 *    - lastgrid: coarse grid (nodes have P0 held in 'B_' - hack)
 *    - right: vector for right scaling
 *
 *  SIDE EFFECTS:
 *    - call 'SetBsz()' for 'SELECTED' nodes
 *
 *  RETURN:
 *    - PET error code 
 */ 
int PromAggGrid::QRShapes( const PromAggGrid *lastgrid ) const
{
#define PROM_TBUFF_SZ (8*4096)
  const int nextmy0 = proc_gnode_[masterProc(comm_.mype())];
  int ierr,todo,tt,ii,jj,xx,kk=PROM_MAXNDF,LWORK=kk*kk,LDA,INFO,proc;
  int bsz = PROM_TBUFF_SZ; 
  double TAU[PROM_MAXNDF+1], WORK[PROM_MAXNDF*PROM_MAXNDF];
  double buff__[PROM_TBUFF_SZ], *pa, *buf = buff__;
  xx = sizeof(shpfloat); jj = sizeof(MPI_Request);  ii=sizeof(int);
  const int fisz=xx/ii, reqsz = (jj/xx + ((jj%xx == 0) ? 0 : 1)) * fisz;
  MPI_Status status; const int tag = lastgrid->getNewTag();
  PromList<int*> *list; PromList<MPI_Request*> smessages; 
  PromTable<PromList<int*>*> proc_list(59);
  PromAggNode *fine, *n1, *curr;

  // QR - copy into shapcgrepe, mark covered
  shpfloat *R = NULL;
  for( xx = 0, todo = nLocalNd_ ; xx < lastgrid->nLocalNd_ ; xx++ ) {
    fine = dynamic_cast<PromAggNode*>(lastgrid->nodes_[xx]);
    if( fine->getState() == PROM_SELECTED ) {
      const int cgid = fine->parents_->data_[0]->global(),clid=cgid - nextmy0;
      const int fndf=fine->getNDF(),bs=fine->getBsz(),cndf=bs/fndf;
      const int cndf2=cndf*cndf;
      shpfloat*Q=fine->parents_->data_[0]->P0_;assert(bs%fndf==0);assert(bs>0);
      // count M
      for(n1=fine,jj=0;n1!=NULL;n1=dynamic_cast<PromAggNode*>(n1->getChild())){
	assert(n1->isMarked()); 
	int Rsz=n1->getBsz(),ndf=n1->getNDF(); assert(Rsz%ndf==0&&Rsz/ndf==cndf);
	shpfloat *vv = n1->parents_->data_[0]->P0_; assert(vv!=NULL);
	for( ii = kk = 0 ; ii < ndf ; ii++ ){
	  for( int j=0 ; j < cndf ; j++, kk++ ) vv[kk] = n1->B_[kk];
	}
	jj += ndf;
	if(ndf!=fndf)SETERRQ4(1,"gid %d:ndf=%d;gid %d:fndf=%d",
			      n1->index_.global(),ndf,fine->index_.global(),
			      fndf);
      }
      const int M = jj;

      // check buffer
      if( M*cndf >= bsz ){ // leave room for sentinal
	if( buf != buff__ ) PetscFree(buf);
	bsz = 4*bsz; if( M*cndf > bsz ) bsz = M*cndf;
	ierr = PetscMalloc(bsz*sizeof(double),&buf); CHKERRQ(ierr);
      }
      LDA = M;
      //
      curr = NULL;
      if( clid >= 0 && clid < nLocalNd_ && isActive() ) { // local
	curr = dynamic_cast<PromAggNode*>(nodes_[clid]); 
	if( curr->B_ == NULL ){ ierr = curr->SetBsz(cndf2); CHKERRQ(ierr); }
	else assert(curr->getBsz() == cndf2);
	R = curr->B_;
	todo--; // have it
      }
      else { // ghost or inactive 
	ierr = GetProc( cgid, proc ); CHKERRQ(ierr); 
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<int*>(10); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	const int ssz = 2 + fisz * cndf2; int *pbf; 
	ierr = PetscMalloc(ssz*sizeof(int),&pbf); int *pb = pbf; CHKERRQ(ierr);
	list->addTail( pbf );
	// < cgid, flid, R[] >
	*pb++ = cgid; *pb++ = xx; 
	R = (shpfloat*)pb;
      }
      // BCs
      for( n1 = fine, jj=0 ; 
	   n1 != NULL ; 
	   n1 = dynamic_cast<PromAggNode*>(n1->getChild()) ) {
	int ndf = n1->getNDF(); 
	for( ii = 0 ; ii < ndf ; ii++ ){
	  if( n1->isFixed( ii ) ) {
	    for( int j=0 ; j < cndf ; j++, jj++ ) Q[jj] = 0.0;
	  }
	  else jj += cndf;
	}
      }
      assert(jj==M*cndf);
      // QR: fine->parents_->data_[0] * curr[0] <= fine->parents_->data_[0]
      assert(Q[M*cndf]==-6.);      assert(R != NULL);
      if( M < cndf ) {
	if( options_.verbose_ > 0 ) {
	  PetscPrintf(PETSC_COMM_SELF,
		      "[%d]%s Warning M(%d) < N(%d), fix %p\n",
		      comm_.mype(),__FUNCT__,M,cndf,curr); 
	}
	assert(M==3 && cndf==6); // only handle this now, R = I, Q = P
	for( ii = 0; ii < 6 ; ii++ ){
	  for( jj = 0; jj < 6 ; jj++ ){
	    // copy up: R = [ I 0 ]
	    if( ii == jj ) {
	      if( ii < 3 ) R[ii*cndf + jj] = 1.;
	      else R[ii*cndf + jj] = 0.0;
	    }
	    else R[ii*cndf + jj] = 0.; // ortho Q
	  }
	}
      }
      else {
	pa = buf;
	if( cndf == 6 ) { // transpose: A = Q'
 	  for( ii = jj = 0 ; ii < M ; ii++, jj += cndf ){
	    pa[ii] = Q[jj];           pa[1*M + ii] = Q[jj + 1];
	    pa[2*M + ii] = Q[jj + 2]; pa[3*M + ii] = Q[jj + 3];
	    pa[4*M + ii] = Q[jj + 4]; pa[5*M + ii] = Q[jj + 5];
	  }
	}
	else {
	  for( ii = jj = 0 ; ii < M ; ii++ ){
	    for( int k = 0 ; k < cndf ; k++, jj++ ) pa[k*M + ii] = Q[jj];
	  }
	}
	// QR
	ii = cndf; kk = M;
	dgeqrf_( &kk, &ii, pa, &LDA, TAU, WORK, &LWORK, &INFO );
	if( INFO < 0 ) SETERRQ(INFO,"xGEQRS error");
	// R
	for( ii = kk = 0; ii < cndf ; ii++ ) { 
	  for( jj = 0; jj < cndf ; jj++, kk++ ) {    
	    if( ii <= jj ) R[kk] = pa[jj*M + ii];
	    else R[kk] = 0.;
	  }
	}
	// Q
	ii = cndf; kk = M;
	dorgqr_( &kk, &ii, &ii, pa, &LDA, TAU, WORK, &LWORK, &INFO );
	if( INFO < 0 ) SETERRQ(INFO,"xORGQR error");
	if( cndf == 6 ) { // transpose: Q = A'
	  for( ii = jj = 0 ; ii < M ; ii++, jj += 6 ){
	    Q[jj] = pa[ii];           Q[jj + 1] = pa[1*M + ii];
	    Q[jj + 2] = pa[2*M + ii]; Q[jj + 3] = pa[3*M + ii];
	    Q[jj + 4] = pa[4*M + ii]; Q[jj + 5] = pa[5*M + ii];
	  }
	}
	else {
	  for( ii = jj = 0 ; ii < M ; ii++, jj += cndf ){
	    for( int k = 0 ; k < cndf ; k++ ) Q[jj + k] = pa[k*M + ii];
	  }
	}
      }
      assert(Q[M*cndf]==-6.);
    }
  }
  if( buf != buff__ ) PetscFree(buf); buf = NULL; bsz = 0;

  // send coarse grid Bs
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    Position pos = list->getHeadPosition(); assert(pos!=NULL);  tt = 0; 
    while( pos ) {
      int *pb2 = list->getNext(pos), xx = pb2[1];
      PromAggNode *fine = dynamic_cast<PromAggNode*>(lastgrid->nodes_[xx]);
      int fndf = fine->getNDF(), cndf = fine->getBsz()/fndf, cndf2 = cndf*cndf;
      int ssz = 2 + fisz*cndf2;                                assert(cndf2>0);
      tt += ssz;
    }
    MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz)*sizeof(int), &sreq ); 
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    assert((size_t)pb%sizeof(shpfloat) == 0);
    while( !list->isEmpty() ) {
      int *pb2 = list->removeHead(), xx = pb2[1];
      PromAggNode *fine = dynamic_cast<PromAggNode*>(lastgrid->nodes_[xx]);
      int fndf = fine->getNDF(), cndf = fine->getBsz()/fndf, cndf2 = cndf*cndf;
      int ssz = 2 + fisz*cndf2; pb2[1] = cndf;// < cgidI, cndf, R >
      PetscMemcpy( pb, pb2, ssz*sizeof(int) ); pb += ssz;
      PetscFree( pb2 );
    }
    assert(tt == pb-buff); 
    // send(tag) *< gidI, cndf2, R >
    MPI_Isend(buff,tt,MPI_INT,proc/lastgrid->factor_,tag,lastgrid->gComm_,sreq);
    delete list;
  }
  // recv Bs 
  if( !isActive() ) todo = 0;
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, lastgrid->gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    proc = status.MPI_SOURCE * lastgrid->factor_;
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag) *< gidI, cndf, R >
    MPI_Recv(buff,tt,MPI_INT,status.MPI_SOURCE,tag,lastgrid->gComm_,&status);
    while( pb-buff < tt ) {
      todo--; assert(todo>=0);
      int cgid = pb[0], cndf = pb[1], cndf2 = cndf*cndf, ssz = 2 + fisz*cndf2; 
      ii = cgid - nextmy0;      assert( ii >= 0 && ii < nLocalNd_ );
      PromAggNode *fine = dynamic_cast<PromAggNode*>(nodes_[ii]);
      R = (shpfloat*)&pb[2];
      if( fine->B_ == NULL ){ ierr = fine->SetBsz(cndf2); CHKERRQ(ierr); }
      for( ii = 0; ii < cndf2 ; ii++ ) fine->B_[ii] = R[ii];
      pb += ssz;
    }
    assert(pb-buff == tt);   PetscFree(buff); 
  }

  // free
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }

  // send next 
  if(options_.verbose_>1)PetscPrintf(lastgrid->gComm_,
				      "[%d]%s SendQRShapes(%d)\n",
				      comm_.mype(),__FUNCT__,getLevel());
  ierr = SendQRShapes( lastgrid ); CHKERRQ(ierr);
  MPI_Iprobe(MPI_ANY_SOURCE, tag, lastgrid->gComm_, &ii, &status);assert(!ii);

  return 0;
}

/* PromAggGrid::SendQRShapes *****************************************
 *
 *  - Cosrse grid method: send - orthigonalize (Q) last grids rigid body modes
 *    of marked nodes
 *
 *  INPUT:
 *    - lastG: grid
 *
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *    - PET cerror code 
 */ 
#undef __FUNCT__ 
#define __FUNCT__ "PromAggGrid::SendQRShapes" 
int PromAggGrid::SendQRShapes( const PromAggGrid *lastG ) const 
{ 
  shpfloat      temp;   
  int           ierr,todo,gid,proc,tt,Esz;PromElement::GetWrite_size(Esz);
  int           ii=sizeof(int),xx=sizeof(temp); tt = sizeof(MPI_Request);
  const int     fisz = xx/ii, reqsz = (tt/xx + ((tt%xx == 0) ? 0 : 1)) * fisz;
  const int     lastmy0 = lastG->proc_gnode_[comm_.mype()];
  MPI_Status    status;                            assert(lastG->isActive()); 
  PromNode *fine;   PromGlobalID *dep;             assert(Esz==2);
  PromList<PromNode*> *list; const int tag = lastG->getNewTag();
  PromList<MPI_Request*> smessages; 
  PromTable<PromList<PromNode*>*> proc_list(59);

  // count uncovered (for and parent data P0 on fine grid)  
  for( xx = 0, todo = 0 ; xx < lastG->nNodes_  ; xx++ ){
    fine = lastG->nodes_[xx];
    if( !fine->isMarked() ) todo++;
  }
  // collect covered 
  for( xx = 0 ; xx < lastG->nNodes_  ; xx++ ){
    fine = lastG->nodes_[xx];
    if( fine->isMarked() && fine->symbData_->dependents != NULL ){
      Position pos = fine->symbData_->dependents->getHeadPosition(); 
      while( pos ) { // send to old ghosts
	dep = fine->symbData_->dependents->getNext( pos );  
	proc = dep->proc();//lastG->factor_; 
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<PromNode*>(100); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->addTail( fine );
      }
    }
  }
  // send covered parent[0]
  while( !proc_list.isEmpty() ){
    list = proc_list.removeHead( &proc ); proc--;
    tt = 0; // list->getCount()*(3+Esz); 
    Position pos = list->getHeadPosition(); assert(pos!=NULL);
    while( pos ) {
      fine = list->getNext(pos);
      tt += 4 + Esz; // 
      int Qsz = fine->parents_->data_[0]->getPsz(); assert(Qsz>0);
      assert( tt%fisz == 0 ); // tt++;
      tt += Qsz*fisz;
    }
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      fine = list->removeHead(); 
      // send back parents < gid, cgid, geq, Psz, PromElement, P0_[] >
      *pb++ = fine->index_.global();PromParent *par = fine->parents_->data_[0];
      *pb++ = par->global();  *pb++ = par->geq(); 
      int Qsz = par->getPsz(); *pb++ = Qsz; par->PromElement::Write( &pb );
      shpfloat *vp = par->P0_; assert(vp!=NULL);
      assert( (pb-buff)%fisz == 0 );
      for(ii = 0 ; ii < Qsz ; ii++){temp = *vp++; PromContext::Write(temp,pb);}
    }
    assert(tt == pb-buff);
    // send(tag) *<gid, cgid, geq, Psz, PromElement, P0_[]> 
    MPI_Isend(buff,pb-buff,MPI_INT,proc/lastG->factor_,tag,lastG->gComm_,
	      sreq);
    delete list; 
  }
  // recieve uncovered parent_[0]
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, lastG->gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); 
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag) *< gid, cgid, geq, Psz, PromElement, P0_[] >
    MPI_Recv(buff,tt,MPI_INT,status.MPI_SOURCE,tag,lastG->gComm_,&status);
    while( pb-buff < tt ) {
      todo--;                                                 assert(todo>=0);
      gid = *pb++; ii = gid - lastmy0; 
      if(ii >= 0 && ii < lastG->nLocalNd_) fine = lastG->nodes_[ii];
      else fine = dynamic_cast<PromAggNode*>(lastG->ghosts_.find(gid+1)); 
      assert(fine!=NULL);                     assert(fine->nParents() == 1); 
      PromParent *par = fine->parents_->data_[0];
      gid = *pb++; assert(gid == par->global()); // parent gid
      par->setGEQ( *pb++ );
      int Qsz = *pb++; //par->getPsz();
      par->PromElement::Read( &pb ); // copy bits (!shp_sz)
      int ndf = fine->getNDF(), cndf = par->getNDF(); assert(Qsz==ndf*cndf);
      if( par->getPsz() == 0 ){  // no repart. allows 0 
	ierr = par->SetPsz( Qsz ); CHKERRQ(ierr);
      }
      else assert(Qsz == par->getPsz());
      if( (pb-buff)%fisz != 0 ){int tv = *pb++; assert(tv==-3);}
      if( Qsz > 1 ) {
	par->SetPsz( Qsz );
	shpfloat *vp = par->P0_; assert(vp!=NULL);
	for( ii=0; ii < ndf ; ii++ ){
	  for( int j = 0 ; j < cndf ; j++ ) {
	    PromContext::Read(temp,pb);
	    if( fine->isFixed(ii) ) *vp++ = 0.0; 
	    else *vp++ = temp;  
	  }
	}
      }
      else{ 
	PromContext::Read(temp,pb);
	if( fine->isFixed(0) ) temp = 0.0;
	ierr = par->setScalarP0( temp ); CHKERRQ(ierr);
      }
      
    }
    assert(tt == pb-buff);   PetscFree(buff);
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 
  // debug
  MPI_Iprobe(MPI_ANY_SOURCE, tag, lastG->gComm_, &ii, &status);assert(!ii);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::MakeInitAdjacs"
/* PromAggGrid::MakeInitAdjacs ******************************
 *
 *  - make approx adjacentcies on coarse nodes. Do not make ghost--ghost.
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromAggGrid::MakeInitAdjacs()
{
  // set adjacs with local cover (are exact w/o smoothing), collect recv  
  for( int xx = 0, nnd = prevgrid_->nLocalNd_ ; xx < nnd ; xx++ ){
    PromNode *curr, *f2, *c2, *fine = prevgrid_->nodes_[xx];
    curr = fine->getParent();            assert(curr!=NULL);
    TablePos tpos = fine->symbData_->adjac.getHeadPosition();
    while( tpos ) {
      f2 = fine->symbData_->adjac.getNext( tpos ); 
      c2 = f2->getParent(); assert(c2!=NULL);
      if( c2 != curr ){
	if( curr->symbData_->adjac.add(c2) ) c2->symbData_->adjac.add(curr);
      }
    }
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::PrepForSolve"
/* PromAggGrid::PrepForSolve ******************************
 *
 */
int PromAggGrid::PrepForSolve()
{
  int ierr;
  
  // make ASM data
  if( !isTop() ) {
    assert(blockISs_ == NULL); bool flag;
    options_.HasOptionName("-aggmg_use_aggragate_blocks", &flag);
    if( flag ){
      ierr = MakeBlocksFromAggragates(); CHKERRQ(ierr);
      if(!(nBlocks_==nextgrid_->nLocalNd_ || factor_!=nextgrid_->factor_))
	PetscPrintf(MPI_COMM_SELF,"\t[%d]%s WARNING: nBlocks_=%d nextgrid_->nLocalNd_=%d nextgrid_->factor_=%d factor_=%d\n",comm_.mype(), __FUNCT__,nBlocks_,nextgrid_->nLocalNd_,nextgrid_->factor_,factor_);
    }
  }
  // cal comm stuff
  ierr = PromGrid::PrepForSolve();  CHKERRQ(ierr);
 
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::MakeBlocksFromAggragates" 
/*  PromAggGrid::MakeBlocksFromAggragates ***********************************
 *
 *   - setup blocks for Schwarz methods
 *
 *   INPUT:
 *    - symbG: other version of this grid
 *
 *   SIDE EFFECTS:
 *    - creates "blockISs"
 * 
 *   RETURN:
 *     - PETSc error code
 */
int PromAggGrid::MakeBlocksFromAggragates()
{
  // make blocks
  nBlocks_ = 0;
  if( isActive() ) {
    int        ierr,xx,bid,ii,ndone,blockID,minblock,nn,*bids;
    PromNode   *sfine, *f2;   assert( blockISs_ == NULL ); // make blockISs
    PromIS      isblock;      assert(sizeof(PromIS) == sizeof(PromStack*));
    PromList<PromIS> list; 
    const int fact=stiffness_->isComplex() ? 2 : 1;
    int nneq = nlocalEq(), buffsz = 3*nLocalNd_ + fact*nneq;
    ierr = PetscMalloc((buffsz+1)*sizeof(int),&bids); CHKERRQ(ierr);
    int *counts=bids+nLocalNd_, *pptrs=bids+2*nLocalNd_, *eqs=bids+3*nLocalNd_;
    bids[buffsz] = -8;
    
    // init blocks
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ) bids[xx] = -1;
    for( xx = ndone = blockID = 0 ; xx < nLocalNd_ ; xx++ ) {
      sfine = nodes_[xx]; 
      if( sfine->getState() == PROM_SELECTED ) {
	assert( !sfine->isGhost() ); 
	for( f2 = sfine, nn = 0 ; f2 ; 
	     f2 = dynamic_cast<PromNode*>(f2->getChild()) ){
	  if( !f2->isGhost() ) { // keep local
	    assert( !f2->isMarked() ); f2->mark(); ndone++; nn++;
	    bids[f2->index_()] = blockID;
	  }
	}
	counts[blockID] = nn; assert(nn>0);
	blockID++;
      }
    }
    if( blockID == 0 ) { // no selected here
      for( xx = 0 ; xx < nLocalNd_ ; xx++ ){ ndone++; bids[xx] = 0; }
      blockID++; // nBlocks
    }
    nBlocks_ = blockID;
    // add straglers
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
      if( bids[xx] == -1 ) {
	sfine = nodes_[xx]; minblock = 32000; nn = -1; 
	TablePos tpos = sfine->symbData_->adjac.getHeadPosition();  
	while( tpos ) { 
	  PromNode *f1 = sfine->symbData_->adjac.getNext( tpos ); 
	  if( !f1->isGhost() && (ii=bids[f1->index_()]) != -1 ) {
	    if( counts[ii] < minblock ) { minblock = counts[ii]; nn = ii; };
	  }
	}
	if( nn == -1 ) { // hard search
	  TablePos tpos = sfine->symbData_->adjac.getHeadPosition();  
	  while( tpos && nn == -1 ) { 
	    PromNode *f1 = sfine->symbData_->adjac.getNext( tpos ); 
	    if( !f1->isGhost() ) {
	      TablePos tpos2 = f1->symbData_->adjac.getHeadPosition();  
	      while( tpos2 && nn == -1 ) { 
		PromNode *f2 = f1->symbData_->adjac.getNext(tpos2);
		if( !f2->isGhost() && (ii=bids[f2->index_()]) != -1 ) nn = ii;
	      }
	    }
	  }
	}
	if( nn == -1 ) nn = 0; // give up
	bids[xx] = nn; 
	counts[nn]++;
	ndone++;
      }
    }
    assert(ndone == nLocalNd_);
    // block lists
    int *bheadlist = counts; counts = NULL;
    for( bid = 0 ; bid < nBlocks_ ; bid++ ) bheadlist[bid] = -1; // head list
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ) { 
      bid = bids[xx];
      pptrs[xx] = bheadlist[bid]; bheadlist[bid] = xx; 
    }
    // make blocks
    for( bid = ndone = 0 ; bid < nBlocks_ ; bid++ ) {
      for( xx = bheadlist[bid], ii = 0 ; xx != -1 ; xx = pptrs[xx], ndone++ ) {
	PromAggNode *fine = dynamic_cast<PromAggNode*>(nodes_[xx]);
	int ndf = fine->getNDF()*fact, eq = fine->geq()*fact;
	for( int j = 0 ; j < ndf ; j++ ) eqs[ii++] = eq++;
      }
      qsort( (void*)eqs, ii, sizeof(int), prom_intcompar );
      ierr = PromISCreateGeneral(PETSC_COMM_SELF,ii,eqs,&isblock);
      CHKERRQ(ierr);
      list.addTail( isblock );
    }
    assert(ndone == nLocalNd_);
    // make IS
    assert(nBlocks_ == list.getCount());
    ierr = PetscMalloc( nBlocks_ * sizeof(PromIS), &blockISs_ );CHKERRQ(ierr); 
    for(xx=0 ; !list.isEmpty() ; xx++) blockISs_[xx] = list.removeHead();
    assert(bids[buffsz] == -8);    PetscFree( bids ); 
    // unmark symbG
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ) nodes_[xx]->mark(FALSE);
  }
  if( nBlocks_ == 0 ) {
    nBlocks_ = 1; int ierr;
    ierr = PetscMalloc(nBlocks_*sizeof(PromIS),&blockISs_); CHKERRQ(ierr);
    ierr = PromISCreateGeneral(PETSC_COMM_SELF, 0, NULL,blockISs_);
    CHKERRQ(ierr);
  }
  
  return 0;
}
